@@ -2,6 +2,7 @@
 #define __CR_PSTREE_H__
 
 #include "common/list.h"
+#include "common/lock.h"
 #include "pid.h"
 #include "images/core.pb-c.h"
 
@@ -15,7 +16,7 @@ struct pstree_item {
 	struct list_head	children;	/* list of my children */
 	struct list_head	sibling;	/* linkage in my parent's children list */
 
-	struct pid		pid;
+	struct pid		*pid;
 	pid_t			pgid;
 	pid_t			sid;
 	pid_t			born_sid;
@@ -24,8 +25,17 @@ struct pstree_item {
 	struct pid		*threads;	/* array of threads */
 	CoreEntry		**core;
 	TaskKobjIdsEntry	*ids;
+	union {
+		futex_t		task_st;
+		unsigned long	task_st_le_bits;
+	};
 };
 
+enum {
+	FDS_EVENT_BIT	= 0,
+};
+#define FDS_EVENT (1 << FDS_EVENT_BIT)
+
 struct pstree_item *current;
 
 struct rst_info;
@@ -66,7 +76,7 @@ static inline bool is_alive_state(int state)
 
 static inline bool task_alive(struct pstree_item *i)
 {
-	return is_alive_state(i->pid.state);
+	return is_alive_state(i->pid->state);
 }
 
 extern void free_pstree(struct pstree_item *root_item);
